/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 539 $ $LastChangedDate: 2002-08-25 13:25:13 +0200 (Sun, 25 Aug 2002) $
 *
 * Ksi - Interface to the Ksi intermediate language.
 */

open Ksi;

//typedef string id;

id make_id(string s)
{
	return s;
}

ty ptroff()
{
	return MacroType["ptr_off"];
}

ty uptroff()
{
	return MacroType["uptr_off"];
}

void output_header(Stdio::FILE out)
{
	Stdio::fputs(out, "// generated by *Gont* interface ;)\n" +
		   	  "!version 1\n\n" +
		   	  "[\n");
}


void output_trailer(Stdio::FILE out)
{
	Stdio::fputs(out, "]\n\n// EOF\n");
}

// Now the writer.

union term {
	*[string, <term>List::t] T;		/* (string ...) */
	<term>List::t TList;			/* [...] */
	*[string, term] TComment;
	string TArg;
	*[int, term] TLine;
	<term>List::t TCompound;
	string TTopComment;
}

term one_int(int i)
{
	return TArg[itoa(i)];
}

<term>List::t convert_string(string s)
{
	int max_len = 50;
	string specials = "\r\n\t\"\\", subst = "rnt\"\\";
	char backslash = String::get("\\", 0);
	
	int pos = 0;
	string tmp = String::create(max_len + 3);
	<term>List::t ret = List::Nil;

	void append(char c)
	{
		if (String::contains(specials, c)) {
			int idx = String::index(specials, c);
			char sbst = String::get(subst, idx);
			String::set(tmp, pos++, backslash); 
			String::set(tmp, pos++, sbst);
		} else {
			String::set(tmp, pos++, c); 
		}

		if (pos >= max_len) {
			string ss = "\"" + String::sub(tmp, 0, pos) + "\".";
			ret = List::Cons[TArg[ss], ret];
			pos = 0;
		}
	}

	String::iter(append, s);

	ret = List::Cons[TArg["\"" + String::sub(tmp, 0, pos) + "\""], ret];
	
	return List::rev(ret);
}

term convert_ty(ty t)
{
	int cnt = 0;
	
	switch t {
	case Void:
		return T["void", List::Nil];
	case Bool:
		return T["bool", List::Nil];
	case Ellipsis:
		return T["ellipsis", List::Nil];
	case Poly:
		return T["poly", List::Nil];
	case TyRef[s]:
		return T["tyref", List::cons1(TArg[s])];
	case Int[s]:
		return T["int", List::cons1(one_int(s))];
	case UInt[s]:
		return T["uint", List::cons1(one_int(s))];
	case Float[s]:
		return T["float", List::cons1(one_int(s))];
	case Ptr[x]:
		return T["ptr", List::cons1(convert_ty(x))];
	case Reference[x]:
		return T["reference", List::cons1(convert_ty(x))];
	case Struct[n]:
		return T["struct", List::cons1(TArg[n])];
	case Union[n]:
		return T["union", List::cons1(TArg[n])];
	case MacroType[n]:
		return TArg["." + n];
	case FuncPtr[ret, args]:
		return T["func_ptr", 
			 List::cons2(convert_ty(ret),
			             TList[List::map(convert_ty, args)])];
	}
}

term convert_attr(attr a)
{
	switch a {
	case Aligned[i]:
		return T["aligned", List::cons1(one_int(i))];
	case Section[name]:
		return T["section", convert_string(name)];
	case Packed:
		return T["packed", List::Nil];
	case Local:
		return T["local", List::Nil];
	case Public:
		return T["public", List::Nil];
	case Extern:
		return T["extern", List::Nil];
	case SuppressDebug:
		return T["suppress_debug", List::Nil];
	case ConstFunc:
		return T["const_func", List::Nil];
	case PureFunc:
		return T["pure_func", List::Nil];
	case StaticCtor:
		return T["static_ctor", List::Nil];
	case StaticDtor:
		return T["static_dtor", List::Nil];
	case MallocLike:
		return T["malloc_like", List::Nil];
	case NoReturn:
		return T["no_return", List::Nil];
	case NoLimitStack:
		return T["no_limit_stack", List::Nil];
	case NoCheckMemoryUsage:
		return T["no_check_memory_usage", List::Nil];
	case Artificial:
		return T["artificial", List::Nil];
	case Weak:
		return T["weak", List::Nil];
	case OneOnly:
		return T["one_only", List::Nil];
	case Comdat:
		return T["comdat", List::Nil];
	case ExportName[name]:
		return T["export_name", convert_string(name)];
	case AliasSet[id]:
		return T["alias_set", List::cons1(TArg[id])];
	}
}


string binop_name(binop op)
{
	switch op {
	case Eq: return "eq";
	case Ne: return "ne";
	case Ge: return "ge";
	case Gt: return "gt";
	case Le: return "le";
	case Lt: return "lt";
	case Plus: return "plus";
	case Minus: return "minus";
	case Mult: return "mult";
	case RDiv: return "rdiv";
	case ExactDiv: return "exact_div";
	case RoundDiv: return "round_div";
	case FloorDiv: return "floor_div";
	case CeilDiv: return "ceil_div";
	case TruncDiv: return "trunc_div";
	case TruncMod: return "trunc_mod";
	case RoundMod: return "round_mod";
	case FloorMod: return "floor_mod";
	case CeilMod: return "ceil_mod";
	case BitAnd: return "bit_and";
	case BitIOr: return "bit_ior";
	case BitAndTc: return "bit_andtc";
	case BitXor: return "bit_xor";
	case Max: return "max";
	case Min: return "min";
	case LRotate: return "lrotate";
	case LShift: return "lshift";
	case RRotate: return "rrotate";
	case RShift: return "rshift";
	case PostInc : return "postinc";
	case PreInc : return "preinc";
	case PostDec : return "postdec";
	case PreDec: return "predec";
	case TruthAndIf: return "truth_andif";
	case TruthOrIf: return "truth_orif";
	case TruthAnd: return "truth_and";
	case TruthOr: return "truth_or";
	case TruthXor: return "truth_xor";
	}
}

string unop_name(unop op)
{
	switch op {
        case Negate: return "negate";
        case BitNot: return "bit_not";
        case TruthNot: return "truth_not";
        case Abs: return "abs";
        case Addr: return "addr";
        case RAddr: return "raddr";
        case Ffs: return "ffs";
        case NonLvalue: return "non_lvalue";
	}
}

term make_var(ty t, id n)
{
	return T["var", List::cons2(convert_ty(t), TArg[n])];
}

term convert_expr(expr ex)
{
	term build_call(string name, term fnc, <expr>List::t args)
	{
		let targs = List::map(convert_expr, args) in
		return T[name, List::cons2(fnc, TList[targs])];
	}

	switch ex {
	case IntConst[t, v]:
		return T["const", List::cons2(convert_ty(t), 
					      TArg[Int64::to_string(v, 10)])];
	case FloatConst[t, v]:
		return T["const", List::cons2(convert_ty(t), 
					      TArg[ftoa(v)])];
					      
	case BoolConst[true]:
		return T["const", List::cons2(convert_ty(Bool), TArg["1"])];
		
	case BoolConst[false]:
		return T["const", List::cons2(convert_ty(Bool), TArg["0"])];
	
	case StringConst[s]:
		return T[".string", convert_string(s)];

	case Null[t]:
		return T["const", List::cons2(convert_ty(t), TArg["0"])];

	case Var[t, n]:
		return make_var(t, n);
	
	case StaticVar[t, n]:
		return T["static_var", List::cons2(convert_ty(t), TArg[n])];

	case Bind[e]:
		return T["bind", List::cons1(convert_expr(e))];
	
	case Ref[n]:
		return T["ref", List::cons1(TArg[n])];
		
	case EnumRef[n]:
		return T["enumref", List::cons1(TArg[n])];
		
	case FuncRef[n]:
		return T["func_ref", List::cons1(TArg[n])];
		
	case Modify[e1, e2]:
		return T["modify", List::cons2(convert_expr(e1), 
					       convert_expr(e2))];
					       
	case Component[e, n]:
		return T["component", List::cons2(convert_expr(e), 
						  TArg[n])];
						  
	case Arrow[e, n]:
		return convert_expr(Component[Indirect[e], n]);

	case Return[Nop]:
		return T["return", List::cons1(T["nop", List::Nil])];

	case Return[e]:
		let assign = T["modify", List::cons2(T["result", List::Nil],
			 			     convert_expr(e))] in
		return T["return", List::cons1(assign)];

	case SizeOf[t]:
		return T["size_of", List::cons1(convert_ty(t))];
		
	case Compound[lst]:
		return TCompound[List::map(convert_expr, lst)];
	
	case Label[n]:
		return T["label", List::cons1(TArg[n])];
		
	case Goto[n]:
		return T["goto", List::cons1(TArg[n])];
		
	case Loop[e]:
		return T["loop", List::cons1(convert_expr(e))];
		
	case ExitLoop[e]:
		return T["exit_loop", List::cons1(convert_expr(e))];
		
	case Nop:
		return T["nop", List::Nil];
	
	case Indirect[e]:
		return T["indirect", List::cons1(convert_expr(e))];
		
	case Cond[e1, e2, e3]:
		return T["cond", List::cons3(convert_expr(e1),
					     convert_expr(e2),
					     convert_expr(e3))];
	
	case Call[n, args]:
		return build_call("call", TArg[n], args);
		
	case FCall[n, args]:
		return build_call("fcall", TArg[n], args);
		
	case PtrCall[e, args]:
		return build_call("ptr_call", convert_expr(e), args);
		
	case PtrFCall[e, args]:
		return build_call("ptr_fcall", convert_expr(e), args);
		
	case BuiltIn[n, args]:
		return build_call("builtin", TArg[n], args);
		
	case FBuiltIn[n, args]:
		return build_call("fbuiltin", TArg[n], args);
	
	case Convert[t, e]:
		return T["convert", List::cons2(convert_ty(t), 
						convert_expr(e))];

	case Block[n, e]:
		return T["block", List::cons2(TArg[n], 
					      convert_expr(e))];

	case ExitBlock[n]:
		return T["exit", List::cons1(TArg[n])];
		
	case Binary[op, e1, e2]:
		return T[binop_name(op), List::cons2(convert_expr(e1),
						     convert_expr(e2))];
		
	case Unary[op, e]:
		return T[unop_name(op), List::cons1(convert_expr(e))];
	
	case TryFinally[e1, e2]:
		return T["try_finally", List::cons2(convert_expr(e1),
						    convert_expr(e2))];
	
	case Line[no, e]:
		return TLine[no, convert_expr(e)];
	
	case Comment[s, e]:
		return TComment[s, convert_expr(e)];
	}
}

term convert_attrs(<attr>List::t a)
{
	return TList[List::map(convert_attr, a)];
}

term convert_decl(decl d)
{
	int cnt = 0;
	
	term convert_unnamed_parm(ty tt)
	{
		switch tt {
		case Ellipsis:
			return T["ellipsis", List::Nil];
		case _:
			return make_var(tt, "_" + itoa(cnt++));
		}
	}

	term convert_named_parm(*[ty, id] p)
	{
		let [t, n] = p in
		return make_var(t, n);
	}

	term convert_enum_fld(*[id, int] fld)
	{
		let [n, v] = fld in
		return T["entry", List::cons2(TArg[n], one_int(v))];
	}
	
	switch d {
	case FuncProto[attrs, ret_type, name, arg_types]:
//print_string("here: " + name + ": "+ itoa(List::length(arg_types)) + "\n");
		let args = List::map(convert_unnamed_parm, arg_types) in
		return T["func",
		    List::Cons[convert_attrs(attrs),
		    List::Cons[convert_ty(ret_type),
		    List::Cons[TArg[name],
		    List::Cons[TList[args],
		    List::Cons[T["nil", List::Nil], 
		    List::Nil]]]]]];
		    
	case Func[attrs, ret_type, name, fargs, body]:
		let args = List::map(convert_named_parm, fargs) in
		return T["func",
		    List::Cons[convert_attrs(attrs),
		    List::Cons[convert_ty(ret_type),
		    List::Cons[TArg[name],
		    List::Cons[TList[args],
		    List::Cons[convert_expr(body), 
		    List::Nil]]]]]];
	
	case TopComment[c]:
		return TTopComment[c];
	
	case StructDefProto[n]:
		return T["structdef", List::cons2(TArg[n], 
						  TList[List::Nil])];
						  
	case StructDef[attrs, n, lst]:
		return T["structdef", 
		     List::cons3(convert_attrs(attrs), TArg[n], 
			    TList[List::map(convert_named_parm, lst)])];

	case UnionDefProto[n]:
		return T["uniondef", List::cons2(TArg[n], 
						  TList[List::Nil])];
						  
	case UnionDef[attrs, n, lst]:
		return T["uniondef", 
		     List::cons3(convert_attrs(attrs), TArg[n],
			    TList[List::map(convert_named_parm, lst)])];
	
	case TyAlias[attrs, n, t]:
		return T["tyalias", List::cons2(convert_attrs(attrs),
						make_var(t, n))];
						
	case AbstractType[attrs, n]:
		return T["abstract_type", List::cons2(convert_attrs(attrs),
			    			TArg[n])];
						
	case EnumDef[lst]:
		return T["enumdef", 
		     List::cons1(TList[List::map(convert_enum_fld, lst)])];
			    
	case Global[attrs, t, n]:
		return T["global", List::cons2(convert_attrs(attrs),
						make_var(t, n))];
						
	case InputFile[s]:
		return T["input_file", convert_string(s)];
	}
}

term strip_tlines(term t, int current_line)
{
	term sak(term tt, int l)
	{
		switch strip_tlines(tt, l) {
		case TCompound[List::Nil]:
			return T["nop", List::Nil];
			
		case TCompound[lst]:
			return T["compound", List::cons1(TList[lst])];

		case x:
			return x;
		}
	}
	
	term self(term tt)
	{
		return sak(tt, current_line);
	}
	
	term raw_self(term tt)
	{
		return strip_tlines(tt, current_line);
	}
	
	switch t {
	case T[n, args]:
		return T[n, List::map(self, args)];
	
	case TList[args]:
		return TList[List::map(self, args)];
	
	case TComment[s, tt]:
		return TComment[s, self(tt)];
	
	case TArg[_]:
		return t;
	
	case TTopComment[_]:
		return t;
	
	case TLine[line, tt]:
		if (line == current_line) {
			return raw_self(tt);
		} else {
			return T["wfl", 
				 List::cons2(one_int(line), sak(tt, line))];
		}
	
	case TCompound[lst]:
		<term>List::t flatten(<term>List::t l)
		{
			switch l {
			case List::Cons[TCompound[hd], tl]:
				return List::concat(hd, flatten(tl));
			case List::Cons[T["nop", _], tl]:
				return flatten(tl);
			case List::Cons[hd, tl]:
				return List::Cons[hd, flatten(tl)];
			case List::Nil:
				return List::Nil;
			}
		}
	
		let l = flatten(List::map(raw_self, lst)) in
		switch l {
		case List::Cons[el, List::Nil]:
			return el;
		case _:
			return TCompound[l];
		}
	}
}

type raw_lterm;

struct lterm {
	raw_lterm raw;
	int len;
}

union raw_lterm {
	*[string, <lterm>List::t] L;
	<lterm>List::t LList;			/* [...] */
	*[string, lterm] LComment;
	string LArg;
	string LTopComment;
}

int infty;

lterm make_lterm(term t)
{
	lterm r;
	<lterm>List::t lst = List::Nil;
	
	switch t {
	case T[n, args]:
		lst = List::map(make_lterm, args);
		r = ({raw = L[n, lst], len = String::length(n) + 2});
	
	case TList[args]:
		lst = List::map(make_lterm, args);
		r = ({raw = LList[lst], len = 2});
	
	case TComment[s, tt]:
		r = ({raw = LComment[s, make_lterm(tt)], len = infty});
	
	case TArg[n]:
		r = ({raw = LArg[n], len = String::length(n)});
	
	case TTopComment[s]:
		r = ({raw = LTopComment[s], len = infty});
	
	case TLine[_, _]:
		raise Sorry["impossible #1"];
		
	case TCompound[_]:
		raise Sorry["impossible #1a"];
	}

	
	let sum = (fun int (int sofar, lterm lt) (sofar + lt.len + 1)) in
	r.len = r.len + List::fold_left(sum, 0, lst);

	return r;
}

void print_lterm_one_line(Stdio::FILE out, lterm lt)
{
	int no = 0;
	
	void self(lterm l)
	{
		if (no++ != 0)
			Stdio::fputs(out, " ");
		print_lterm_one_line(out, l);
	}
	
	switch lt.raw {
	case L[n, lst]:
		Stdio::fputs(out, "(" + n);
		no++;
		List::iter(self, lst);
		Stdio::fputs(out, ")");
		
	case LList[lst]:
		Stdio::fputs(out, "[");
		List::iter(self, lst);
		Stdio::fputs(out, "]");
	
	case LArg[s]:
		Stdio::fputs(out, s);
	
	case x:
		raise Sorry["impossible #2"];
	}
	
}

void print_lterm(Stdio::FILE out, int indent, lterm lt)
{
	char space = String::get(" ", 0);
	bool sparse = false;

	void print_deeper(lterm l)
	{
		print_lterm(out, indent + 1, l);
	}
	
	if (lt.len < infty) {
		Stdio::fputs(out, String::make(indent, space));
		print_lterm_one_line(out, lt);
		if (!sparse)
			Stdio::fputs(out, "\n");
		return;
	}
	
	switch lt.raw {
	case LTopComment[s]:
		Stdio::fputs(out, String::make(indent, space) + 
				  "// " + s + "\n");
		
	case LComment[s, t]:
		Stdio::fputs(out, String::make(indent, space) + 
				  "// " + s + "\n");
		print_lterm(out, indent, t);

	case L[n, lst]:
		string ind = String::make(indent, space);
		
		Stdio::fputs(out, ind + "(" + n + "\n");
		List::iter(print_deeper, lst);
		if (sparse)
			Stdio::fputs(out, ")");
		else
			Stdio::fputs(out, ind + ")\n");
		
	case LList[lst]:
		string ind = String::make(indent, space);
		
		if (sparse) {
			Stdio::fputs(out, ind + "[");
			List::iter(print_deeper, lst);
			Stdio::fputs(out, "]");
		} else {
			Stdio::fputs(out, ind + "[\n");
			List::iter(print_deeper, lst);
			Stdio::fputs(out, ind + "]\n");
		}
	
	case LArg[_]:
		Stdio::fputs(out, String::make(indent, space));
		print_lterm_one_line(out, lt);
		if (!sparse)
			Stdio::fputs(out, "\n");
	}
}

bool fast_writer;

void output_decl(Stdio::FILE out, decl d)
{
	print_lterm(out, 0, make_lterm(strip_tlines(convert_decl(d), -1)));
}

section init 
{
	fast_writer = false;
	infty = 75;
}
