/* 
 * Copyright (c) 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 539 $ $LastChangedDate: 2002-08-25 13:25:13 +0200 (Sun, 25 Aug 2002) $
 *
 * Runtime environment for Gont, Std module implementation.
 */

open Std;

/* Just wrappers... */
extern_c void Std__exit(int);

/* ... and very simple functions... */
'a exit(int r)
{
	Std__exit(r); 
	raise Exit; 	// never reached
}

/* Stdout output. */
void print_int(int i)
{
	print_string(itoa(i)); 
}

void print_float(float f)
{
	print_string(ftoa(f)); 
}

void print_newline()
{
	print_string("\n");
}

void print_endline(string s)
{
	print_string(s);
	print_newline();
}

/* Stderr output. */
void prerr_int(int i)
{
	prerr_string(itoa(i)); 
}

void prerr_float(float f)
{
	prerr_string(ftoa(f)); 
}

void prerr_newline()
{
	prerr_string("\n");
}

void prerr_endline(string s)
{
	prerr_string(s);
	prerr_newline();
}

/* Input. */
int read_int()
{
	return atoi(read_line());
}

float read_float()
{
	return atof(read_line());
}

/* Raise shortcuts. */
'a fail_with(string s)
{
	raise Failure[s];
}

'a invalid_argument(string s)
{
	raise Invalid_argument[s];
}

/* Pair opertations. */
'a fst(*['a, 'b] p)
{
	switch p {
	case [a, _]: return a;
	}
}

'b snd(*['a, 'b] p)
{
	switch p {
	case [_, a]: return a;
	}
}

void __dummy_disaster()
{
	// make sure these are linked (other are referenced elsewhere in
	// this module)
	raise Null_access;
	raise Match_failure;
	raise Not_found;
	raise End_of_file;
	raise Sys_error[""];
}

/* Character type. */
typedef int char;

char char_of_int(int x)
{
	if (x < 0 || x > 255)
		raise Invalid_argument["char_of_int"];
	return x;
}

int int_of_char(char x)
{
	return x;
}
